use crate::{Code, Instructions, Object, SymbolTable, SymbolScope};
use tonic_parser::{Statement, Expression, Op};

#[derive(Debug)]
pub struct Compiler {
    instructions: Instructions,
    constants: Vec<Object>,

    symbols: SymbolTable,

    scopes: Vec<CompilationScope>,
    scope: usize,
}

#[derive(Debug, Default)]
struct CompilationScope {
    instructions: Instructions,
}

impl Compiler {
    
    pub fn new() -> Self {
        Self {
            instructions: Vec::new(),
            constants: Vec::new(),

            symbols: SymbolTable::new(),

            scopes: vec![
                CompilationScope::default(),
            ],
            scope: 0,
        }
    }

    fn current_scope(&mut self) -> &mut CompilationScope {
        self.scopes.get_mut(self.scope).unwrap()
    }

    pub fn instructions(&mut self) -> Vec<Code> {
        self.current_scope().instructions.clone()
    }

    pub fn compile(&mut self, statement: Statement) {
        match statement {
            Statement::Expression { expression } => {
                self.expression(expression);
                self.emit(Code::Pop);
            },
            Statement::If { condition, then, otherwise } => {
                self.expression(condition);

                // Emit a JumpIfNot with a fake instruction pointer.
                let jump_if_not_position = self.emit(Code::JumpIfNot(9999));

                for statement in then {
                    self.compile(statement);
                }

                if otherwise.is_empty() {
                    let after_then_position = self.current_scope().instructions.len();

                    self.replace(jump_if_not_position, Code::JumpIfNot(after_then_position));
                } else {
                    let jump_position = self.emit(Code::Jump(9999));

                    let after_then_position = self.current_scope().instructions.len();
                    self.replace(jump_if_not_position, Code::JumpIfNot(after_then_position));

                    for statement in otherwise {
                        self.compile(statement);
                    }

                    let after_otherwise_position = self.current_scope().instructions.len();
                    self.replace(jump_position, Code::Jump(after_otherwise_position));
                }
            },
            Statement::Let { identifier, initial, .. } => {
                self.expression(initial);

                let symbol = self.symbols.define(identifier);
                let code = if symbol.scope == SymbolScope::Global {
                    Code::SetGlobal(symbol.index)
                } else {
                    Code::SetLocal(symbol.index)
                };

                self.emit(code);
            },
            Statement::Function { identifier, parameters, body, .. } => {
                self.enter_scope();

                let parameter_count = parameters.len();

                for parameter in parameters {
                    self.symbols.define(parameter.name);
                }

                for statement in body {
                    self.compile(statement);
                }

                // Add a return to the end of the function to enforce a `null` return value.
                self.emit(Code::Return(false));

                let instructions = self.leave_scope();

                let function = Object::Function(instructions, self.symbols.count, parameter_count);
                let constant = self.constant(function);

                self.emit(Code::Constant(constant));

                let symbol = self.symbols.define(identifier);
                let global = Code::SetGlobal(symbol.index);
                
                self.emit(global);
            },
            Statement::Return { expression } => {
                self.expression(expression);

                self.emit(Code::Return(true));
            },
            _ => todo!()
        }
    }

    pub fn expression(&mut self, expression: Expression) {
        match expression {
            Expression::Infix(left, op, right) => {
                self.expression(*left);
                self.expression(*right);

                match op {
                    Op::Add => self.emit(Code::Add),
                    Op::Subtract => self.emit(Code::Subtract),
                    Op::Multiply => self.emit(Code::Mulitply),
                    Op::Divide => self.emit(Code::Divide),
                    Op::GreaterThan => self.emit(Code::GreaterThan),
                    Op::LessThan => self.emit(Code::LessThan),
                    Op::GreaterThanEquals => self.emit(Code::GreaterThanEquals),
                    Op::LessThanEquals => self.emit(Code::LessThanEquals),
                    Op::Equals => self.emit(Code::Equals),
                    Op::NotEquals => self.emit(Code::NotEquals),
                    _ => todo!()
                };
            },
            Expression::Number(n) => {
                let position = self.constant(Object::Number(n));

                self.emit(Code::Constant(position));
            },
            Expression::String(s) => {
                let position = self.constant(Object::String(s));

                self.emit(Code::Constant(position));
            },
            Expression::Bool(b) => {
                if b { self.emit(Code::True) } else { self.emit(Code::False) };
            },
            Expression::Prefix(op, right) => {
                self.expression(*right);

                match op {
                    Op::Subtract => self.emit(Code::Negate),
                    _ => todo!()
                };
            },
            Expression::Identifier(i) => {
                let symbol = self.symbols.resolve(&i).clone();

                if symbol.is_none() {
                    panic!("undefined variable: {}", i);
                }

                let symbol = symbol.unwrap();
                let code = if symbol.scope == SymbolScope::Global {
                    Code::GetGlobal(symbol.index)
                } else {
                    Code::GetLocal(symbol.index)
                };

                self.emit(code);
            },
            Expression::Call(function, arguments) => {
                self.expression(*function);

                let count = arguments.len();

                for arg in arguments {
                    self.expression(arg);
                }

                self.emit(Code::Call(count));
            },
            _ => todo!("{:?}", expression),
        }
    }

    fn replace(&mut self, position: usize, code: Code) {
        self.current_scope().instructions[position] = code;
    }

    fn constant(&mut self, object: Object) -> usize {
        self.constants.push(object);
        self.constants.len() - 1
    }

    fn emit(&mut self, code: Code) -> usize {
        self.current_scope().instructions.push(code);

        self.current_scope().instructions.len() - 1
    }

    fn enter_scope(&mut self) {
        let scope = CompilationScope::default();

        self.scopes.push(scope);
        self.scope += 1;

        self.symbols = SymbolTable::from_outer(self.symbols.clone());
    }

    fn leave_scope(&mut self) -> Instructions {
        let scope = self.scopes.pop().unwrap();

        self.scope -= 1;
        self.symbols = self.symbols.outer();

        scope.instructions
    }
}

#[derive(Debug)]
pub struct Bytecode {
    pub instructions: Instructions,
    pub constants: Vec<Object>,
}

impl From<Compiler> for Bytecode {
    fn from(mut c: Compiler) -> Self {
        Self { instructions: c.current_scope().instructions.clone(), constants: c.constants }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use tonic_parser::{Lexer, Parser};

    #[test]
    fn infixes() {
        assert_eq!(compile("1 + 1;"), vec![
            Code::Constant(0),
            Code::Constant(1),
            Code::Add,
            Code::Pop,
        ]);
    }

    #[test]
    fn prefixes() {
        assert_eq!(compile("-1;"), vec![
            Code::Constant(0),
            Code::Negate,
            Code::Pop
        ]);
    }

    #[test]
    fn bools() {
        assert_eq!(compile("true; false;"), vec![
            Code::True,
            Code::Pop,
            Code::False,
            Code::Pop,
        ]);
    }

    #[test]
    fn strings() {
        assert_eq!(compile(r##""foo";"##), vec![
            Code::Constant(0),
            Code::Pop,
        ]);

        assert_eq!(compile(r##""foo" + "bar";"##), vec![
            Code::Constant(0),
            Code::Constant(1),
            Code::Add,
            Code::Pop,
        ]);
    }

    #[test]
    fn functions() {
        assert_eq!(compile("return 5;"), vec![
            Code::Constant(0),
            Code::Return(true),
        ]);

        assert_eq!(compile("fn foo() { 5 + 10; }"), vec![
            Code::Constant(2),
            Code::SetGlobal(0),
        ]);

        assert_eq!(compile("fn foo() { return 5 + 10; }"), vec![
            Code::Constant(2),
            Code::SetGlobal(0),
        ]);

        assert_eq!(compile("fn foo() {} foo();"), vec![
            Code::Constant(0),
            Code::SetGlobal(0),
            Code::GetGlobal(0),
            Code::Call(0),
            Code::Pop,
        ]);

        assert_eq!(compile("fn foo(bar) { return bar; }"), vec![
            Code::Constant(0),
            Code::SetGlobal(0),
        ]);
    }

    #[test]
    fn conditionals() {
        assert_eq!(compile("if true { 10; } 3333;"), vec![
            Code::True,
            Code::JumpIfNot(4),
            Code::Constant(0),
            Code::Pop,
            Code::Constant(1),
            Code::Pop,
        ]);

        assert_eq!(compile("if false { 10; } 3333;"), vec![
            Code::False,
            Code::JumpIfNot(4),
            Code::Constant(0),
            Code::Pop,
            Code::Constant(1),
            Code::Pop,
        ]);

        assert_eq!(compile("if true { 10; } else { 20; } 3333;"), vec![
            Code::True,
            Code::JumpIfNot(5),
            Code::Constant(0),
            Code::Pop,
            Code::Jump(7),
            Code::Constant(1),
            Code::Pop,
            Code::Constant(2),
            Code::Pop,
        ]);
    }

    #[test]
    fn globals() {
        assert_eq!(compile("let foo = 1;"), vec![
            Code::Constant(0),
            Code::SetGlobal(0),
        ]);

        assert_eq!(compile("let foo = 1; foo;"), vec![
            Code::Constant(0),
            Code::SetGlobal(0),
            Code::GetGlobal(0),
            Code::Pop,
        ]);
    }

    fn compile(source: &str) -> Vec<Code> {
        let program = parse(source);
        let mut compiler = Compiler::new();

        for statement in program {
            compiler.compile(statement);
        }

        compiler.instructions().clone()
    }

    fn parse(source: &str) -> Vec<Statement> {
        let lexer = Lexer::new(source);
        let mut parser = Parser::new(lexer);

        parser.read();
        parser.read();

        parser.parse()
    }
}